///***************************************************************************************
// BMP280/BME280 Pressure Sensor
// Do I2C Bit Banger
// Uses 1 MHz DCO
// MSP430G2553
// Jed Margolin  11/20/2020.
//***************************************************************************************
#include <msp430.h>
#include <math.h>
#include "bme280.h"
//================================================
// Read i2C byte
unsigned char i2c_read(unsigned char address)
{
    unsigned char ack;

// Make sure SDA is an output and high
//    SDA_PDIR |= SDA_BIT;     //Out
//    SDA_POUT |= SDA_BIT;

// Start bit
    i2c_start_bit();

// send the ID
   ack = i2c_sendbyte(IIC_ID_WRITE);

// we will keep going even if ack is high
   ack = ack | i2c_sendbyte(address);

// How does the slave send the data back?

// Repeat the start bit
   i2c_start_bit();

// Send the device ID
   ack = ack | i2c_sendbyte(IIC_ID_READ);

   ack = ack | i2c_getbyte();     // data ic2_rxdata is global

// Give it a Stop bit
   i2c_stop_bit();

   return(i2c_rxdata);      // global
}
//=====================================
void i2c_start_bit()
{

// Make sure SCL is low since an SDA transition with SCL low will be ignored
   SCL_POUT &= ~SCL_BIT;

// Make SDA high
   SDA_POUT |= SDA_BIT;
   wait(7);

   SCL_POUT |= SCL_BIT;   // SCL high
   wait(5);
   SDA_POUT &= ~SDA_BIT;   //SDA Low
   wait(2);

   SCL_POUT &= ~SCL_BIT;   //SCL low
   wait(2);
}
//===================================================
void i2c_stop_bit()
{

// Make sure SCL is low since an SDA transition with SCL low will be ignored
   SCL_POUT &= ~SCL_BIT;
   wait(4);

// Make SDA low
   SDA_POUT &= ~SDA_BIT;
   wait(8);

// Make SCL high
   SCL_POUT |= SCL_BIT;   // SCL high

   wait(8);
   SDA_POUT |= SDA_BIT;   //SDA high

   wait(10);
}
//=================================================
unsigned char i2c_sendbyte(unsigned char data)
{
   unsigned char n, ack;

   for(n=8; n>0; n--)
   {
     if(data & 0x80)
     {
        SDA_POUT |= SDA_BIT;   // SDA high
     }
     else {SDA_POUT &= ~SDA_BIT;}  // SDA low
     data = data<<1;

     wait(6);
     SCL_POUT |= SCL_BIT;   //SCL high
     wait(7);
     SCL_POUT &= ~SCL_BIT;   //SCL low
//       wait(10);
   }

// look for ACK from I2C device
   SDA_PDIR &= ~SDA_BIT;   // Input

   wait(4);
   SCL_POUT |= SCL_BIT;   //SCL high

// ack is immediately available
   if(SDA_PIN & SDA_BIT)
   {
      ack = 1;
   }
   else {ack = 0;}

   wait(8);

// Then make SCL low
   SCL_POUT &= ~SCL_BIT;   //SCL low

// Give the slave time to release SDA
   wait(100);

// Then make SDA an output again and low; SCL is low so transitions won't matter
   SDA_POUT &= ~SDA_BIT;
   SDA_PDIR |= SDA_BIT;     // An output again

   wait(10);

// Then SDA low
//    SDA_POUT &= ~(SDA_BIT);     //SDA low
//    wait(10);

   return(ack);
}
//=======================================
unsigned char i2c_getbyte()
{
   unsigned char n, data;

// Make SDA an input
   SDA_PDIR &= ~SDA_BIT;   // In

   data = 0;

   for(n=8; n>0; n--)
   {
      data = data<<1;
      if(SDA_PIN & SDA_BIT)
      {
         data = data | 0x01;   // SDA high
      }

      wait(6);
      SCL_POUT |= SCL_BIT;   //SCL high
      wait(7);
      SCL_POUT &= ~SCL_BIT;   //SCL low
   }

// We are responsible for sending ACK or NACK
// so make SDA an output
   SDA_PDIR |= SDA_BIT;     //Out/

// An ACK tells the slave to keep sending data, a NACK says no more
// We are only asking for one byte here so send a NACK
   SDA_POUT |= SDA_BIT;

   wait(4);
   SCL_POUT |= SCL_BIT;   //SCL high
   wait(10);
   SCL_POUT &= ~SCL_BIT;   //SCL low

   SDA_POUT |= SDA_BIT;  // SDA back to high to get ready for the stop bit

// and we are done
   i2c_rxdata = data;
   return(0);
}
//====================================
int wait_ready()
{
    int n, timeout;
    unsigned char busy, data;

    busy = 1;
    n = 1000;
    timeout = 0;

    while(busy)
    {
       data = i2c_read(0xF3);
       if( !(data & 0x08)) {busy = 0;}
       if(n == 0)  {timeout = 1;}
       n--;
    }
    return timeout;
}
//==================================
void wait(int i)
{
   unsigned int n;

   for(n=i; n>0; n--)
   {
     asm(" nop");
   }
}
//=======================================================
void bme_init(void)
{
    unsigned char address;

// BME280 reset
    address = bme_reset;
    i2c_write(address,0);

    wait1();

// Initialize the BME-280
// config    d7   d6   d5 | d4   d3   d2 | d1 | d0
//              t_sb          filter        x   spi3W_en
//            0    0    0    0    0    0    0    0
// t_sb = 000 (0.5 ms  standby time
// filter = 000  filter off
// spi3W_en = 0 for 4-wire
    address = bme_config;
    i2c_write(address,0x00);

// humidity: oversampling x1   ('0' = sensor off)
    i2c_write(0xF2,0x01);

// ctrl_meas    d7   d6   d5 |  d4   d3   d2 | d1   d0
//                 osrs_t           osrs_p       mode
//               0    0    1     0    1    1   0     1
// osrs_t = 001  oversampling x 1 17-bit  ('0' = sensor off)
// osrs_p = 011 oversampling x4   ('0' = off)
// mode = 11 (normal)  (0 = sleep; 01 and 10 = Forced Mode)
   address = bme_ctrl_meas;
   i2c_write(address,0x2D);  // In forced mode it must be called for every reading
   wait1();
   wait_ready();
}
//==============================================
// Get the calibration values
void getcal()
{
    unsigned char rxtemp;
    int k;

    rxtemp = 0x88;
    for(k=0; k<28; k++)
    {
        wait_ready();
        caldata[k]= i2c_read(rxtemp++);
    }

// Temperature

// Make digs for compensation functions. All digs are global
//Requires that the calata array be unsigned
// caldata[0] = dig_T1 lsb  caldata[1] = dig_T1 msb unsigned
   dig_T1 = caldata[1]<<8 | caldata[0];

// caldata[2] = dig_T2 lsb  caldata[3] = dig_T2 msb signed
   dig_T2 = caldata[3]<<8 | caldata[2];

// caldata[4] = dig_T3 lsb  caldata[5] = dig_T3 msb signed
   dig_T3 = caldata[5]<<8 | caldata[4];


// Pressure
// caldata[6] = dig_P1 lsb  caldata[7] = dig_P1 msb unsigned
   dig_P1 = caldata[7]<<8 | caldata[6];

// caldata[8] = dig_P2 lsb  caldata[9] = dig_P2 msb signed
   dig_P2 = caldata[9]<<8 | caldata[8];
//   if(dig_P2 & 0x10000) { dig_P2 |= 0xFFFF0000;}  // make signed

// caldata[10] = dig_P3 lsb caldata[11] = dig_P3 msb    signed
   dig_P3 = caldata[11]<<8 | caldata[10];

// caldata[12] = dig_P4 lsb caldata[13] = dig_P4 msb    signed
   dig_P4 = caldata[13]<<8 | caldata[12];

// caldata[14] = dig_P5 lsb caldata[15] = dig_P5 msb    signed
   dig_P5 = caldata[15]<<8 | caldata[14];

// caldata[16] = dig_P6 lsb caldata[17] = dig_P6 msb    signed
   dig_P6 = caldata[17]<<8 | caldata[16];

// caldata[18] = dig_P7 lsb caldata[19] = dig_P7 msb    signed
   dig_P7 = caldata[19]<<8 | caldata[18];

// caldata[20] = dig_P8 lsb caldata[21] = dig_P8 msb    signed
   dig_P8 = caldata[21]<<8 | caldata[20];

// caldata[22] = dig_P9 lsb caldata[23] = dig_P9 msb    signed
   dig_P9 = caldata[23]<<8 | caldata[22];

// Humidity
   wait_ready();
   humdata[0]= i2c_read(0xA1);   // for dig_H1 (unsigned char)

   wait_ready();
   humdata[1]= i2c_read(0xE1);   // for dig_H2 7:0

   wait_ready();
   humdata[2]= i2c_read(0xE2);   // for dig_H2 15:8

   wait_ready();
   humdata[3]= i2c_read(0xE3);   // for dig_H3 7:0  (unsigned char)

   wait_ready();
   humdata[4]= i2c_read(0xE4);   // for dig_H4 11:4

   wait_ready();
   humdata[5]= i2c_read(0xE5);   // 3:0 for dig_H4; 7:4 for for dig_H5 3:0

   wait_ready();
   humdata[6]= i2c_read(0xE6);   // for dig_H5 11:4

   wait_ready();
   humdata[7]= i2c_read(0xE7);   // for dig_H6 (signed char)

// make the digs
   // humdata[0] is dig_H3 7:0 (unsigned char);
   dig_H1 = humdata[0];  //  7:0

   // humdata[1] is dig_H2 7:0; humdata[2] is dig_H2 8:15  (result is )
   dig_H2 = (humdata[2]<<8) | humdata[1];

   // humdata[3] is dig_H3 7:0  (unsigned char);
   dig_H3 = humdata[3];   // dig_H3 7:0

   // humdata[4] (7:0) is dig_H4 11"4; humdata[5] (3:0) is dig_H4 3:0
   dig_H4 = (humdata[4] <<4) | (humdata[5] & 0x0F);

   // humdata[5] 7:4 is also dig_H5 3:0; humdata[6] is dig_H5 11:4
   dig_H5 = (humdata[5]>>4) | (humdata[6]<<4);

   dig_H6 = humdata[6];     // signed char
}
//=====================================================
// Read the BME280 data for pressure, temperature, and humidity
void read_bme280(void)
{
   unsigned int k;
   unsigned long temp1, temp2;
   unsigned char rxtemp,address;

   address = bme_ctrl_meas;
   i2c_write(address,0x2D);  // In forced mode it must be called for every reading
   wait5ms();

   rxtemp = press_msb;         // press_msb  bme_id  calib00 = 0x88
   wait_ready();
   for(k=0; k<6; k++)
   {
      wait_ready();
      bmedata[k]= i2c_read(rxtemp++);
   }

   bmedata[2] &= 0x0F;     // press_msb, pressure is maximum of 20 bits
   bmedata[5] &= 0x0F;     // temp_msb, temperature is maximum of 20 bits

//       adc_p = bmedata[0]<<16 | bmedata[1]<<8 | bmedata[2];  // 0 - msb, 1 = lsb, 2 = xlsb
   adc_p = bmedata[0];
   adc_p = adc_p<<8;
   adc_p = adc_p | bmedata[1];
   adc_p = adc_p <<8;
   adc_p = adc_p | bmedata[2];
   adc_p = adc_p >>4;

//       bmedata[5] = 0x20;       // for testing
   adc_t = 0;               // 3 - msb, 4 = lsb, 5 = xlsb
   adc_t = bmedata[3];     // for 19-bit resolution
   adc_t = adc_t<<8;        // 8
   adc_t = adc_t | bmedata[4];
   adc_t = adc_t<<8;        // 3
   adc_t = adc_t | bmedata[5];
   adc_t = adc_t>>4;

// Humidity
   temp1 = ( (unsigned long)i2c_read(0xFD)) <<8;       // msb
   temp2 = (unsigned long)i2c_read(0xFE);
   adc_h = temp1 | temp2;

//       address = bme_ctrl_meas;     // force mode for next time
//       spi_writeword(address,0x04);        // 0x33

   do_compensation();       // F is temperature in Fahrenheit; p is pressure in mb; humidity is in %
   asm(" nop");    // for debugger
}
//==========================================================
// Compensation for Temperature, Pressure, and Humidity
int do_compensation()
{
    long var1, var2;
    long var3, var4, var5;

    long t_fine;
/*
    long adc_T = 519888;    // uncompensated temperature
    unsigned short dig_T1 = 27504;
    short dig_T2 = 26435;
    short dig_T3 = -1000;

    unsigned long adc_P = 415148;   // uncompensated pressure
    unsigned short dig_P1 = 36477;
    short dig_P2 = -10685;
    short dig_P3 = 3024;
    short dig_P4 = 2855;
    short dig_P5 = 140;
    short dig_P6 = -7;
    short dig_P7 = 15500;
    short dig_P8 = -14600;
    short dig_P9 = 6000;
*/

// Temperature Compensation
    var1 = ((((adc_t>>3) - ((long)dig_T1<<1))) * ((long)dig_T2)) >> 11;
    // var1 = 128793

    var2 = (((((adc_t>>4) - ((long)dig_T1)) * ((adc_t>>4) - ((long)dig_T1))) >> 12) * ((long)dig_T3)) >> 14;
    // var2 = -371

    t_fine = var1 + var2;
    // t_ fine = 128422

    T = (t_fine * 5 + 128) >> 8;
    // T = 2508 = 25.08 deg C

    F = T * 9 / 5 + 3200;
    // F = 7714

// Pressure Compensation
// Returns pressure in Pa as unsigned 32 bit integer. Output value of "96386" equals 96386 Pa = 963.86 hPa
    var1 = (((long)t_fine)>>1) - (long)64000;
    // var = 211

    var2 = (((var1>>2) * (var1>>2)) >> 11 ) * ((long)dig_P6);
    // var2 = -7

    var2 = var2 + ((var1*((long)dig_P5))<<1);
    // var2 = 59073

    var2 = (var2>>2)+(((long)dig_P4)<<16);
    // var2 = 187120048

    var1 = (((dig_P3 * (((var1>>2) * (var1>>2)) >> 13 )) >> 3) + ((((long)dig_P2) * var1)>>1))>>18;
    // var1 = -5

    var1 =((((32768+var1))*((long)dig_P1))>>15);
    // var1 = 36471

    if (var1 == 0)
    {
        return 0; // avoid exception caused by division by zero
    }

    p = (((unsigned long)(((long)1048576) - adc_p) - (var2>>12))) * 3125;
    // p = 1836703125

    if (p < 0x80000000)
    {
       p = (p << 1) / ((unsigned long)var1);
    }
    else
    {
       p = (p / (unsigned long)var1) * 2;
    }
    // p = 100721

    var1 = (((long)dig_P9) * ((long)(((p>>3) * (p>>3))>>13)))>>12;
    // var1 = 28343

    var2 = (((long)(p>>2)) * ((long)dig_P8))>>13;
    // var2 = -44877

    p = (unsigned long)((long)p + ((var1 + var2 + dig_P7) >> 4));
    // p = 100656

// Humidity Compensation
// Bosch code
//  Returns humidity in %RH as unsigned 32 bit integer in Q22.10 format (22 integer and 10 fractional bits).
//  Output value of 47445" represents 47445/1024 = 46.333 %RH

// From the datasheet
/*
    var = (t_fine - ( (signed long)76800) );

    var = ( ( ( ( (adc_h << 14) - ( ( (signed long)dig_H4) << 20) - ( ( (signed long)dig_H5) *
       var))  + ( (signed long)16384) )  >> 15) * ( ( ( ( ( ( (var  *
       ( (signed long)dig_H6)) >> 10)  *  ( ( (var * ( (signed long)dig_H3)) >> 11) +
       ( (signed long)32768) )) >> 10)  +  ((signed long)2097152)) * ((signed long)dig_H2) +
       8192) >> 14));

    var = (var  - ( ( ( ( (var >> 15) * (var >> 15)) >> 7) *  ((signed long)dig_H1) ) >> 4));
    var = (var < 0 ? 0 : var);
    var = (var > 419430400 ? 419430400 : var);

    Humidity = ((unsigned long)var>>12)/1000 ;  // "46332" is 46.332 %
*/

// Bosch Code from https://github.com/BoschSensortec/BME280_driver
// The results are the same as the previous code from the data sheet
    var1 = t_fine - (long)76800;
    var2 = (long)adc_h * 16384;           // adc_h is the uncompensated humidity
    var3 = (long)dig_H4 * 1048576;
    var4 = (long)dig_H5 * var1;
    var5 = ( ((var2 - var3) - var4) + (long)16384 ) / 32768;
    var2 = (var1 * (long)dig_H6 ) / 1024;
    var3 = ( (var1 * (long)dig_H3) ) / 2048;
    var4 = ( (var2 * (var3 + (long)32768)) / 1024) + (long)2097152;
    var2 = ( (var4 * (long)dig_H2 ) + 8192) / 16384;
    var3 = var5 * var2;
    var4 = ((var3 / 32768) * (var3 / 32768)) / 128;
    var5 = var3 - ((var4 * (long)dig_H1 ) / 16);
    var5 = (var5 < 0 ? 0 : var5);
    var5 = (var5 > 419430400 ? 419430400 : var5);
    Humidity = ( (unsigned long)var5 / 4096);

    // mine
    Humidity3 = Humidity/100;
    Humidity = Humidity/1000;

    if (Humidity > 99)
    {
       Humidity = 99;
    }

    if (Humidity3 > 999)
    {
       Humidity3 = 999;
    }
    return 1;
}
//==========================================================
void wait1()
{
volatile unsigned int i;

    i = 0;
//    for(i=0; i<1; i++){}
    return;
}
//=================================================================
// Write to a register with i2c
unsigned char i2c_write(unsigned char address, unsigned char data)
{
    unsigned char ack;

// Make sure SDA is an output and high
//   SDA_PDIR |= SDA_BIT;     //Out
//   SDA_POUT |= SDA_BIT;

// Start bit
     i2c_start_bit();

// send the ID
    ack = i2c_sendbyte(IIC_ID_WRITE);

// we will keep going even if ack is high
    ack = ack | i2c_sendbyte(address);

// Send the data
    ack = ack | i2c_sendbyte(data);

 // Give it a Stop bit
    i2c_stop_bit();

    return(ack);
 }
//==========================================================
